/*
  device_scsi_sg - VPD support routines for SCSI devices using SG interface.
  
  This file is part of the Linux lsvpd package.
  
  (C) Copyright IBM Corp. 2005, 2006
  
  Maintained by  Martin Schwenke <martins@au.ibm.com>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2, or (at your option)
  any later version.
 
  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  
  $Id: device_scsi_sg.c,v 1.1 2006/04/11 18:38:28 emunson Exp $
*/

/* For asprintf(3). */
#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <string.h>
#include <regex.h>
#include <scsi/sg_cmds.h>
#include <unistd.h>

#include "lsvpd_lib_dir.h"
#include "util.h"
#include "vpd.h"
#include "lsvpd.h"
#include "init.h"
#include "sysfs.h"
#include "proc_scsi.h"

#include "device.h"
#include "device_scsi.h"
#include "device_ide.h"

#define DEVICE_SCSI_SG_TEMPLATE_FILE "scsivpd.conf"

/********************************************************************/

/*
 * DS field handling.
 *
 * A certain amount of this is to make the output the same as AIX
 * lsvpd.
 */

#define SCSI_CDROM_DEFAULT "Other SCSI CD-ROM Drive"
static const struct string_array cdrom_device_types[] =
{
	{0, SCSI_CDROM_DEFAULT},
	{1, "OEM SCSI CD-ROM Drive"},
	{2, "SCSI CD-ROM Drive"},
	{3, "SCSI Multimedia CD-ROM Drive"},
	{4, "OEM SCSI DVD-ROM Drive"},
	{5, "SCSI DVD-ROM Drive"},
	{6, "SCSI DVD-RAM Drive"},
	{7, "OEM SCSI DVD-RAM Drive"},
	{-1, NULL}
};

static char *
make_ds_value(const node_t type,
	      const unsigned char * data,
	      unsigned char subtype,
	      const char * dir)
{
	char * ret;
	unsigned char * data_0xc7;
	size_t len_0xc7;
	char * s;
	char * l;
	int cdtype;

	ret = device_scsi_ds(type, subtype);

	/* Only do the fancy stuff for real SCSI devices. */
	if ((NULL != ret) &&
	    ((0 == strcmp(type, DEVICE_TYPE_SCSI)) ||
	     (0 == strcmp(type, DEVICE_TYPE_RAID)))) {
		/* Get info for naming quirks, such as LVD, Multimedia. */
		lsvpd_file_relative_read(dir, "0xc7", &data_0xc7, &len_0xc7);

		/* Check for SCSI *Multimedia* CD-ROM Drive. */
		if (subtype == 5) {
			cdtype = ((NULL != data_0xc7) &&
				  (34 <= len_0xc7)) ?
				(int) data_0xc7[33] : 0;
			s = strdup(string_array_lookup(cdrom_device_types,
						       SCSI_CDROM_DEFAULT,
						       cdtype));
		} else {
			l = ((0 == subtype) && (NULL != data_0xc7) &&
			     (35 <= len_0xc7) && (7 == data_0xc7[34])) ?
				"LVD " : "";
			if (-1 == asprintf(&s, "%s%s", l , ret))
				s = NULL;
		}

		/* Free previous guess. */
		free(ret);
		ret = NULL;
			
		if (NULL != s) {
			if (-1 == asprintf(&ret, "%s%s",
					   ((data[7] & 0x40) ? "32 Bit " :
					    ((data[7] & 0x20) ? "16 Bit " : "")),
					   s))
				ret = NULL;
			free(s);
		}

		if (NULL != data_0xc7)
			free(data_0xc7);
	}
	
	return ret;
}

/********************************************************************/

/* Template handling. */

typedef struct {
	char * v_t_m;
	char * templates;
} get_templates_t;

/* Check if len bytes of template s are a regexp matching string vtm. */
static bool
template_matches(const char * s,
		 size_t len,
		 const char * vtm)
{
	bool ret;
	char * r;
	regex_t preg;

	ret = false;

	/* Got to copy since arg is const. */
	if (NULL != (r = malloc(len + 1))) {
		memcpy(r, s, len);
		r[len] = '\0';
		
		/* Is this the one? */
		if (0 == regcomp(&preg, r, REG_EXTENDED|REG_NOSUB)) {
			ret = (0 == regexec(&preg, vtm, 0, NULL, 0));
			regfree(&preg);
		}
		free(r);
	}

	return ret;
}

static bool
get_templates_line_scanner(const char * line,
			   void * data)
{
	bool ret;
	const char * t;
	int i;
	get_templates_t * inout;

	/* Default is we need to look at subsequent lines. */
	ret = true;
	inout = data;
	inout->templates = NULL;

	/* Check for comments. */
	if ((0 < strlen(line)) && ('#' != line[0])) {
		/* Find the 3rd semicolon. */
		i = 0;
		t = line;
		while ((i < 3) && (NULL != (t = strchr(t, ';')))) {
			i++;
			t++; /* Move past semi-colon. */
		}

		if ((3 == i) &&
		    (template_matches(line, t - 1 - line, inout->v_t_m))) {
			inout->templates = strdup(t);
			ret = false;
		}
	}

	return ret;
}

static char *
get_templates(const char * vendor,
	      const char * type,
	      const char * model)
{
	char * ret;
	char * template_file;
	get_templates_t inout;

	ret = NULL;

	template_file = lsvpd_relative_filename(lsvpd_lib_dir,
						DEVICE_SCSI_SG_TEMPLATE_FILE);
	if (NULL != template_file) {
		if (-1 != asprintf(&inout.v_t_m,
				   "%s;%s;%s", vendor, type, model)) {
			if (line_scanner_harness(template_file,
						 get_templates_line_scanner,
						 &inout))
				ret = inout.templates;
			free(inout.v_t_m);
		}
		free(template_file);
	}

	return ret;
}

/* Like strchr, but replace char by '\0' and return pointer to next char. */
static char *
strchr_nul_plus_plus(char * s, char c)
{
	char * ret;

	ret = strchr(s, c);
	if (NULL != ret) {
		*ret = '\0';
		ret++;
	}

	return ret;
}

static void
process_template(char *        template,
		 const char *  tmp_dir,
		 const char *  subtype_string,
		 vpd_list_t ** vpd)
{
	unsigned char * buf;
	size_t length;
	int offset;
	char * s;
	char * k;
	char * v;
	int field_len;

	/* NUL-terminate page name. */
	s = template;
	s = strchr_nul_plus_plus(s, '=');

	/* Can't read file?  No problem: template might be too specific. */
	if ((NULL != s) &&
	    (lsvpd_file_relative_read(tmp_dir, template, &buf, &length))) {

		/*
		 * Templates skip 1st 32 bytes of standard inquiry,
		 * since this is standard and handled elsewhere.
		 */
		offset = (strcmp(template, "inquiry") == 0) ? 32 : 0;

		while ((NULL != s) && (offset < length)) {
			/* Get field key. */
			k = s;
			s = strchr_nul_plus_plus(s, ':');
			if (NULL != s) {
				/* Get field value length. */
				v = s;
				s = strchr_nul_plus_plus(s, ',');
				field_len = atoi(v);

				/*
				 * Dump value to file if desired and
				 * have enough bytes in buf.
				 */
				if ((k[0] != '_') &&
				    (offset + field_len <= length)){
					device_scsi_append_field(vpd,
								 k,
								 buf + offset,
								 field_len,
								 subtype_string);
				}

				/* Done, move along. */
				offset += field_len;
			}
		}
		free(buf);
	}
}

static void
process_templates(char *        templates,
		  const char *  tmp_dir,
		  const char *  subtype_string,
		  vpd_list_t ** vpd)
{
	char * s;
	char * t;

	s = templates;

	while (NULL != s) {
		t = s;

		s = strchr_nul_plus_plus(s, ';');
		process_template(t, tmp_dir, subtype_string, vpd);
	}
}

/********************************************************************/

static bool
device_scsi_sg_render(const node_t type,
		      const char *   tmp_dir,
		      const char *   subtype_file,
		      vpd_list_t **  vpd)
{
	bool ret;
	char * mf;
	char * tm;
	unsigned char * data;
	size_t len;
	unsigned char subtype;
	const char * subtype_string;
	char * ds;
	char * templates;

	ret = false;
	
	/* Grab contents of inquiry file. */
	if (!lsvpd_file_relative_read(tmp_dir, "inquiry", &data, &len))
		goto done;

	/*
	 * Validate data: if the inquiry data is short or it tells us
	 * that the "target is not capable of supporting a device on
	 * this logical unit, then don't bother proceeding.
	 */
	if ((len < 32) || (0x7F == data[0]))
		goto free_data;

	subtype = data[0] & 0x1F;
	subtype_string = device_scsi_subtype_string(subtype);

	mf = strndup_trim(data + 8, 8);
	tm = strndup_trim(data + 16, 16);

	/* Now do DS and type. */
	ds = make_ds_value(type, data, subtype, tmp_dir);
	if (NULL == ds)
		goto free_fields;

	/* Parse out standard fields.  Need the values later. */
	device_scsi_append_field(vpd, "Z0", data, 8, subtype_string);

	/* "type" is used externally. */
	if (lsvpd_file_write_string(subtype_file, subtype_string)) {
		ret = true;

		if ((NULL != tm) && (NULL != mf)) {
			/* Parse out extra fields using templates. */
			templates = get_templates(mf, subtype_string, tm);
			if (NULL != templates) {
				process_templates(templates, tmp_dir,
						  subtype_string, vpd);
				free(templates);
			}

			device_scsi_ata_mf_tm(type, &mf, &tm);
		}

		device_scsi_append_field_string(vpd, "TM", tm, subtype_string);
		device_scsi_append_field_string(vpd, "MF", mf, subtype_string);
		device_scsi_append_field_string(vpd, "DS", ds, subtype_string);
	}

 free_fields:
	free(mf);
	free(tm);
	free(ds);
 free_data:
	free(data);
 done:
	return ret;
}

/********************************************************************/

/* No need for this buffer to be on the stack. */
static unsigned char device_scsi_sg_buf[4096];

/*
 * Some USB devices lock up if they get anything but an allocation
 * length of 36 for the standard inquiry.
 * 
 * Thanks to Douglas Gilbert for pointing this out to me.
 */
#define DEVICE_SCSI_SG_DEFAULT_STD_LEN 36

/*
 * The allocation length was changed from a 1 byte field to a 2 byte
 * field between SPC-2 and SPC-3.  Therefore, an SPC-2 compliant
 * device may object if a non-zero value is passed in (what is now)
 * the MSB of the allocation length.  Also, some SCSI commands with a
 * fixed response size round down to the nearest multiple of 4 bytes,
 * hence 252 instead of 255.
 *
 * Thanks again to Douglas Gilbert for answering my questions about
 * this.
 */
#define DEVICE_SCSI_SG_DEFAULT_EVPD_LEN 252

/* A conservation - 1. */
#define DEVICE_SCSI_SG_MAX_LEN (sizeof(device_scsi_sg_buf) - 1)

static size_t
device_scsi_sg_resp_len(bool evpd)
{
	
	/*
	 * SPC-3 (7.6) defines some EVPD pages with the page length in
	 * bytes 2 and 3, and some that just use byte 3 with byte 2
	 * reserved.  However, 3.3.9 says:
	 *
	 *   [...] A reserved bit, byte, word or field shall be set to
	 *   zero, or in accordance with a future extension to this
	 *   standard. Recipients are not required to check reserved
	 *   bits, bytes, words or fields for zero values. Receipt of
	 *   reserved code values in defined fields shall be reported
	 *   as an error.
	 */
	return evpd ?
		(device_scsi_sg_buf[2] * 256 +
		 device_scsi_sg_buf[3] + 4) :
		(device_scsi_sg_buf[4] + 5);
}

static bool
device_scsi_sg_sanity_check(bool   evpd,
			    int    page_code)
{
	bool ret;
	char last;
	int i;

	/* No sanity checking for non-EVPD pages. */
	if (!evpd)
		return true;

	/* If we asked for a particular EVPD page and we got something
	 * else then we have failed. */
	if (page_code != device_scsi_sg_buf[1])
		return false;

	/* That's all for EVPD page != 0. */
	if (0 != page_code)
		return true;

	/* Check EVPD page 0 specifies a legal/useful number of page codes... */
	ret = (0 < device_scsi_sg_buf[3]) && (0 == device_scsi_sg_buf[2]);

	/* ...  and the list of available pages is ordered. */
	i = 4;
	last = device_scsi_sg_buf[i];
	i++;
	while (ret && (i < (size_t) device_scsi_sg_buf[3])) {
		ret = (device_scsi_sg_buf[i] > last);
		last = device_scsi_sg_buf[i];
		i++;
	}

	return ret;
}

static bool
device_scsi_sg_ll_inq(int      fd,
		      bool     evpd,
		      int      page_code,
		      size_t * len)
{
	bool ret;
	size_t alloc_len;
	int r;

	ret = false;

	alloc_len = evpd ?
		DEVICE_SCSI_SG_DEFAULT_EVPD_LEN :
		DEVICE_SCSI_SG_DEFAULT_STD_LEN;

	r = sg_ll_inquiry(fd, false, (int) evpd, page_code,
			  device_scsi_sg_buf, (int) alloc_len,
			  false, false); /* Silent! */
	if (0 == r) {
		if (!device_scsi_sg_sanity_check(evpd, page_code))
			return false;

		*len = device_scsi_sg_resp_len(evpd);

		if (*len > alloc_len) {
			alloc_len = *len;
			if (alloc_len > DEVICE_SCSI_SG_MAX_LEN)
				// FIXME: do this, fail or malloc?
				alloc_len = DEVICE_SCSI_SG_MAX_LEN;
			r = sg_ll_inquiry(fd, false, (int) evpd, page_code,
					  device_scsi_sg_buf, (int) alloc_len,
					  false, false);
			if (0 == r) {
				if (!device_scsi_sg_sanity_check(evpd,
								 page_code))
					return false;
				
				*len = device_scsi_sg_resp_len(evpd);
			}
		}
	}

	if (0 == r) {
		// FIXME: do this, or fail?
		if (*len > DEVICE_SCSI_SG_MAX_LEN)
			*len = DEVICE_SCSI_SG_MAX_LEN;
		ret = true;
	}

	return ret;
}

static bool
device_scsi_sg_do_inq(int          fd,
		      bool         evpd,
		      int          page_code,
		      const char * tmp_dir)
{
	bool ret;
	size_t len;
	char * f;

	ret = false;


	if (device_scsi_sg_ll_inq(fd, evpd, page_code, &len)) {
		if (evpd) {
			if (-1 == asprintf(&f, "%s/0x%02x",
					   tmp_dir, (unsigned char) page_code))
				f = NULL;
		} else {
			if (-1 == asprintf(&f, "%s/inquiry", tmp_dir))
				f = NULL;
		}

		if (NULL != f) {
			ret = lsvpd_file_write(f, device_scsi_sg_buf, len);
			free(f);
		}
	}
			
	return ret;
}

static bool
device_scsi_sg_do_inquiries(const node_t type,
			    const bus_info_t bus_info,
			    const char * tmp_dir)
{
	bool ret;
	device_node_info_t * i;
	device_temp_t * d;
	unsigned char *  buf;
	size_t len;
	int j;
	int page_code;

	ret = false;

	i = device_scsi_get_info(type, bus_info);
	if (NULL == i)
		goto done;

	d = device_open_temporary(i);
	if (NULL == d)
		goto free_info;

	ret = device_scsi_sg_do_inq(d->fd, false, 0, tmp_dir);
	if (!ret)
		goto device_close;

	/* EVPD inquiries don't affect the return value. */
	if (device_scsi_sg_do_inq(d->fd, true, 0, tmp_dir) &&
	    (lsvpd_file_relative_read(tmp_dir, "0x00", &buf, &len))) {
		for (j = 0; j < buf[3]; j++) {
			page_code = (int) buf[4+j];
			if (0 != page_code)
				device_scsi_sg_do_inq(d->fd,
						      true, page_code,
						      tmp_dir);
		}
		free(buf);
	}

 device_close:
	device_close_temporary(&d);
 free_info:
	free(i);
 done:
	return ret;
}

/********************************************************************/

static bool
device_scsi_sg_retrieve_vpd(const node_t type,
			    const bus_info_t bus_info,
			    const char * node,
			    vpd_list_t ** vpd)
{
	bool ret;
	char * tmp_dir;
	char * subtype_file;

	ret = false;
	/*
	 * Do all possible inquiries and then render, instead of doing
	 * the EVPD inquiries as required by templates.  This helps to
	 * collect extra inquiry information for debugging and further
	 * development.  If this proves to be a bottleneck, then this
	 * decision can be revisited and they can be retrieved on
	 * demand.
	 */
	subtype_file = lsvpd_relative_filename(node, LSVPD_SUBTYPE_FILE);
	if (NULL != subtype_file) {
		tmp_dir = lsvpd_relative_filename(node, LSVPD_TMP_SUBDIR);
		if (NULL != tmp_dir) {
			ret = lsvpd_mkdir(tmp_dir) &&
				device_scsi_sg_do_inquiries(DEVICE_TYPE_SCSI,
							    bus_info,
							    tmp_dir) &&
				device_scsi_sg_render(type, tmp_dir,
						      subtype_file, vpd);
			free(tmp_dir);
		}
		free(subtype_file);
	}

	if (NULL != vpd)
		device_scsi_vpd_sort(vpd);

	return ret;
}

/********************************************************************/

static struct {
	char **      base;
	const char * rel;
} have_sg_checks[] = {
	{ &lsvpd_sysfs_dir,     "class/scsi_generic" },
	{ &lsvpd_proc_scsi_dir, "sg"                 },
};

static bool
have_sg(void)
{
	bool ret;
	char * t;
	int i;

	ret = false;

	i = 0;
	while (!ret &&
	       ((sizeof(have_sg_checks) / sizeof(have_sg_checks[0])) > i)) {
		if (NULL != *have_sg_checks[i].base) {
			t = lsvpd_relative_filename(*have_sg_checks[i].base,
						    have_sg_checks[i].rel);
			if (NULL != t) {
				ret = (0 == access(t, F_OK));
				free(t);
			}
		}
		i++;
	}

	return ret;
}

static void
init(void)
{
	if (have_sg() ||
	    (system("modprobe sg >/dev/null 2>&1"), have_sg()))
		device_scsi_vpd_function_append(device_scsi_sg_retrieve_vpd);
}

INIT(init);
