/*
  device_scsi_sg_map - VPD support routines for SCSI devices.
  
  This file is part of the Linux lsvpd package.
  
  (C) Copyright IBM Corp. 2002, 2003, 2004, 2005, 2006
  
  Maintained by  Martin Schwenke <martins@au.ibm.com>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2, or (at your option)
  any later version.
 
  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  
  $Id: device_scsi_sg_map.c,v 1.1 2006/04/11 18:38:28 emunson Exp $
*/

/* For asprintf(3). */
#define _GNU_SOURCE

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>

#include "lsvpd_lib_dir.h"
#include "util.h"
#include "vpd.h"
#include "lsvpd.h"
#include "init.h"

#include "device.h"
#include "device_scsi.h"

/********************************************************************/

/*
 * Temporary map created by device_scsi_sg_map_setup (below) and used
 * by other functions.
 */
#define DEVICE_SCSI_SG_MAP "sg_map-x"

/*
 * Full path of sg_map executable. This could be changed to be more
 * flexible, and we only need it on Linux 2.4, so this should do.
 */
#define SG_MAP_PROGRAM "/usr/bin/sg_map"

static bool
device_scsi_sg_map_setup(void)
{
	bool ret;
	int i;
	char node[16];
	char * cmd;

	ret = false;

	// FIXME: This is really unfortunate.  There must be a better
	// way, but sg_map runs against existing device nodes.
	for (i = 0; i < 256; i++)
		if (0 < snprintf(node, sizeof(node), "/dev/sg%d", i))
			mknod(node, 0660 | S_IFCHR, makedev(21, i));

	/* Put together an sg_map command to save the output. */
	if (-1 != asprintf(&cmd,
			   "%s -x > %s/%s 2>/dev/null",
			   SG_MAP_PROGRAM,
			   lsvpd_db_misc_dir,
			   DEVICE_SCSI_SG_MAP)) {
		ret = (0 == system(cmd));
		free(cmd);
	}

	return ret;
}

/********************************************************************/

static bool
device_scsi_sg_map_list_scanner(const char * line,
				void * x)
{
	unsigned int h;
	unsigned int b;
	unsigned int t;
	unsigned int l;
	char * bus_info;
	bus_info_list_t ** node_list;

	node_list = x;

	if (4 == sscanf(line, "%*s %u %u %u %u %*u", &h, &b, &t, &l))
		if (-1 != asprintf(&bus_info, "%s%c%u:%u:%u:%u",
				   DEVICE_TYPE_SCSI, BUS_INFO_SEP,
				   h, b, t, l)) {
			node_list_add(node_list, DEVICE_TYPE_SCSI, bus_info);
			free(bus_info);
		}

	return true;
}

static bool
device_scsi_sg_map_list_devices(bus_info_list_t ** node_list)
{
	bool ret;
	char * map;

	ret = false;

	device_scsi_sg_map_setup();

	map = lsvpd_relative_filename(lsvpd_db_misc_dir,
				      DEVICE_SCSI_SG_MAP);
	if (NULL != map) {
		ret = line_scanner_harness(map,
					   device_scsi_sg_map_list_scanner,
					   node_list);
		free(map);
	}

	return ret;
}

typedef struct {
	device_scsi_address_t sa;
	char * name;

} device_scsi_sg_map_get_name_t;

static bool
device_scsi_sg_map_name_scanner(const char * line,
				void * x)
{
	bool ret;
	unsigned int h;
	unsigned int b;
	unsigned int t;
	unsigned int l;
	char devname[256];
	device_scsi_sg_map_get_name_t * gn;

	gn = x;
	ret = true;

	if ((5 == sscanf(line, "/dev/%*s %u %u %u %u %*u /dev/%255s",
			 &h, &b, &t, &l, devname)) &&
	    ('\0' != devname[0]) && // FIXME: Can't happen?
	    (gn->sa.host   == h) &&
	    (gn->sa.bus    == b) &&
	    (gn->sa.target == t) &&
	    (gn->sa.lun    == l)) {
		gn->name = strdup(devname);
		ret = false;
	} else if ((5 == sscanf(line, "/dev/%255s %u %u %u %u %*u",
				devname, &h, &b, &t, &l)) &&
		   ('\0' != devname[0]) && // FIXME: Can't happen?
		   (gn->sa.host   == h) &&
		   (gn->sa.bus    == b) &&
		   (gn->sa.target == t) &&
		   (gn->sa.lun    == l)) {
		gn->name = strdup(devname);
		ret = false;
	}

	return ret;
}	

static char *
device_scsi_sg_map_get_name (const node_t type,
			     const bus_info_t bus_info)
{
	char * ret;
	device_scsi_sg_map_get_name_t gn;
	char * map;

	ret = NULL;

	if (device_parse_scsi_bus_info(bus_info, &gn.sa)) {

		gn.name = NULL;

		map = lsvpd_relative_filename(lsvpd_db_misc_dir,
					      DEVICE_SCSI_SG_MAP);
		if (NULL != map) {
			if (line_scanner_harness(map,
						 device_scsi_sg_map_name_scanner,
						 &gn))
				ret = gn.name;
			free(map);
		}
	}

	return ret;
}

typedef struct {
	device_scsi_address_t sa;
	bool found;
	unsigned int minor;

} device_scsi_sg_map_get_info_t;

static bool
device_scsi_sg_map_info_scanner(const char * line,
				void * x)
{
	bool ret;
	unsigned int h;
	unsigned int b;
	unsigned int t;
	unsigned int l;
	device_scsi_sg_map_get_info_t * gi;

	gi = x;
	ret = true;

	if ((5 == sscanf(line, "/dev/sg%u %u %u %u %u %*u",
			 &(gi->minor), &h, &b, &t, &l)) &&
	    (gi->sa.host   == h) &&
	    (gi->sa.bus    == b) &&
	    (gi->sa.target == t) &&
	    (gi->sa.lun    == l)) {
		gi->found = true;
		ret = false;
	}

	return ret;
}	

static device_node_info_t *
device_scsi_sg_map_get_info (const node_t type,
			     const bus_info_t bus_info)
{
	device_node_info_t * ret;
	device_scsi_sg_map_get_info_t gi;
	char * map;

	ret = NULL;

	if (device_parse_scsi_bus_info(bus_info, &gi.sa)) {
		gi.found = false;

		map = lsvpd_relative_filename(lsvpd_db_misc_dir,
					      DEVICE_SCSI_SG_MAP);
		if (NULL != map) {
			if ((line_scanner_harness(map,
						  device_scsi_sg_map_info_scanner,
						  &gi)) &&
			    (true == gi.found) &&
			    (NULL != (ret = malloc(sizeof(device_node_info_t))))) {
				ret->major = 21;
				ret->minor = gi.minor;
				ret->mode = S_IFCHR;
			}
			free(map);
		}
	}
	
	return ret;
}

/********************************************************************/

static void
init(void)
{
	node_functions_t s = {
		.type             = DEVICE_TYPE_SCSI,
		.get_kernel_name  = device_scsi_sg_map_get_name,
	};

	/* Only register functions if /usr/bin/sg_map exists. */
	if (0 == access(SG_MAP_PROGRAM, X_OK)) {
		node_type_register(&s);
		node_listing_function_register(device_scsi_sg_map_list_devices);
		device_scsi_get_info_function = device_scsi_sg_map_get_info;
	}
}

INIT(init);
