/*
  vpd_util - VPD support routines.
  
  This file is part of the Linux lsvpd package.
  
  (C) Copyright IBM Corp. 2002, 2003, 2004, 2005, 2006
  
  Maintained by  Martin Schwenke <martins@au.ibm.com>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2, or (at your option)
  any later version.
 
  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  
  $Id: util.c,v 1.1 2006/04/11 18:38:28 emunson Exp $
*/
  
/* For asprintf, get_current_dir_name: */
#define _GNU_SOURCE

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <errno.h>
#include <sys/param.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <ctype.h>
#include <ftw.h>

#include "util.h"

/********************************************************************/

/* Filename manipulation. */

char *
lsvpd_relative_filename(const char * dir,
			const char * relpath)
{
	char * ret;

	if (-1 == asprintf(&ret, "%s/%s", dir, relpath)) {
		ret = NULL;
	}
	return ret;
}

char *
lsvpd_basename(const char * f)
{
	char * ret;
	char * t;

	ret = NULL;

	t= strrchr(f, '/');
	if ((NULL != t) && ('\0' != t[1]))
		ret = strdup(t+1);

	return ret;
}

char *
lsvpd_dirname(const char * f)
{
	char * ret;
	char * t;

	ret = NULL;

	t= strrchr(f, '/');
	if ((NULL != t) && ('\0' != t[1])) {
		ret = malloc(t - f + 1);
		if (NULL != ret) {
			memcpy(ret, f, t - f);
			ret[t - f] = '\0';
		}
	}

	return ret;
}

/********************************************************************/

/* Directory creation and removal. */

bool
lsvpd_mkdir(const char * dir)
{
	bool ret;

	ret = false;

	if ((0 == mkdir(dir, 0777)) || (EEXIST == errno))
		ret = true;
	else if ((ENOENT == errno) &&
		 lsvpd_mkdirname(dir))
		ret = ((0 == mkdir(dir, 0777)) || (EEXIST == errno));

	return ret;
}

bool
lsvpd_mkdirname(const char * f)
{
	bool ret;
	char * t;

	ret = false;

	t = lsvpd_dirname(f);
	if (NULL != t) {
		ret = lsvpd_mkdir(t);
		free(t);
	}

	return ret;
}

bool
lsvpd_mksubdir(const char * dir, const char * subdir)
{
	bool ret;
	char * t;

	ret = false;

	t = lsvpd_relative_filename(dir, subdir);
	if (NULL != t) {
		ret = lsvpd_mkdir(t);
		free(t);
	}

	return ret;
}

static int
lsvpd_rm_rf_unlink(const char * file,
		   const struct stat * sb,
		   int flag,
		   struct FTW * s)
{
	int ret = 0;

	/* Remove anything except "." or "..". */
	if ((0 != strcmp(file, ".")) &&
	    (0 != strcmp(file, "..")))
		ret = unlink(file);

	return ret;
}

void
lsvpd_rm_rf(const char * prefix,
	    const char * d)
{
	if (d != strstr(d, prefix)) {
		/* Get here and you have a serious bug. */
		fprintf(stderr,
			"lsvpd_remove_directory: "
			"internal error \"%s\" not a prefix of \"%s\"\n",
			prefix, d);
		return;
	}
		

	nftw(d,
	     &lsvpd_rm_rf_unlink,
	     64,
	     FTW_DEPTH | FTW_MOUNT| FTW_PHYS);
}

/********************************************************************/

/* File I/O. */

static bool
read_to_eof(FILE   * file,
	    unsigned char ** buf,
	    size_t * length)
{
	size_t n;
	size_t done = 0;

	const size_t chunk_size = 4096;
	int chunk_count = 1;

	if (NULL == (*buf = malloc(chunk_size * chunk_count)))
		goto bail;
	
	/* - 1 everywhere because we're going to bonus NUL-terminate. */
	while (0 < (n = fread(*buf + done, 1,
			      (chunk_size * chunk_count - 1) - done, file))) {
		done += n;
		if (done == (chunk_size * chunk_count - 1)) {
			chunk_count++;
			if (NULL ==
			    (*buf = realloc(*buf, chunk_size * chunk_count)))
				goto bail;
		}
	}
	
	if (0 != ferror(file))
		goto bail;

	(*buf)[done] = '\0';

	if (NULL != length) {
		*length = done;
	}

	return true;

 bail:
	if (NULL != length)
		*length = 0;
	if (NULL != *buf)
		free(*buf);
	*buf = NULL;
	return false;
}

bool
lsvpd_file_read(const char * pathname,
		unsigned char ** buf,
		size_t *     length)
{
	FILE * file;
	char * cmd = NULL;
	bool ret;

	*buf = NULL;

	if (NULL == pathname)
		file = stdin;
	else {
		if (strrchr(pathname, '|') !=
		    (pathname + strlen(pathname) - 1)) {
			if (NULL == (file = fopen(pathname, "r")))
				goto bail;
		} else {
			/* I don't think this is used anywhere. */
			cmd = strdup(pathname);
			cmd[strlen(pathname) - 1] = '\0';
			if (NULL == (file = popen(cmd, "r")))
				goto bail;
		}
	}

	ret = read_to_eof(file, buf, length);

	if (NULL != cmd) {
		free(cmd);
		pclose(file);
	} else
		fclose(file);

	return ret;

 bail:
	if (NULL != length)
		*length = 0;
	if (NULL != cmd)
		free(cmd);
	return false;
}

bool
lsvpd_file_relative_read(const char * dir,
			 const char * relpath,
			 unsigned char ** buf,
			 size_t *     length)
{
	bool ret;
	char * file;

	ret = false;

	if (NULL != (file = lsvpd_relative_filename(dir, relpath))) {
		if (lsvpd_file_read(file, buf, length))
			ret = true;
		free(file);
	}

	return ret;
}

char *
lsvpd_file_read_string(const char * pathname)
{
	unsigned char  * ret;
	size_t  length;

	if (!lsvpd_file_read(pathname, &ret, &length))
		ret = NULL;

	return (char *) ret;
}

char *
lsvpd_file_relative_read_string(const char * dir,
				const char * relpath)
{
	unsigned char  * ret;
	size_t  length;

	if (!lsvpd_file_relative_read(dir, relpath, &ret, &length))
		ret = NULL;
	
	return (char *) ret;
}

bool
lsvpd_file_write(const  char  * pathname,
		 const  unsigned char * buf,
		 size_t  count)
{
	bool ret;
	int fd;
	size_t n;
	size_t done = 0;

	ret = false;

	if (-1 != (fd = open(pathname, O_CREAT|O_WRONLY|O_TRUNC, 0644))) {
		ret = true;
		while (done != count) {
			if (-1 == (n = write(fd, buf + done, count - done))) {
				ret = false;
				break;
			}
				
			done += n;
		}
		
		close(fd);
	}

	return ret;
}

bool
lsvpd_file_relative_write(const char * dir,
			  const char * relpath,
			  const unsigned char * buf,
			  size_t   count)
{
	bool ret;
	char * file;

	ret = false;

	if (NULL != (file = lsvpd_relative_filename(dir, relpath))) {
		if (lsvpd_file_write(file, buf, count))
			ret = true;
		free(file);
	}

	return ret;
}

bool
lsvpd_file_write_string(const char * pathname,
			const char * string)
{
	return lsvpd_file_write(pathname,
				(unsigned char *) string,
				strlen(string));
}

bool
lsvpd_file_relative_write_string(const char * dir,
				 const char * relpath,
				 const char * string)
{
	return lsvpd_file_relative_write(dir, relpath,
					 (unsigned char *) string,
					 strlen(string));
}

/********************************************************************/

/* Miscellaneous. */

char *
lsvpd_readlink(const char * l)
{
	char relval[MAXPATHLEN+1];
	int n;

	n = readlink(l, relval, sizeof(relval) - 1);
	if (-1 == n)
		return NULL;
	relval[n] = '\0';

	return strdup(relval);
}

const char *
string_array_lookup(const struct string_array * a,
		    const char * d,
		    int k)
{
	int i = 0;
	while (a[i].val != NULL) {
		if (a[i].key == k) {
			return a[i].val;
		}
		i++;
	}
	return d;
}


char *
lsvpd_strnmove_trim(char * dest,
		    const unsigned char * src,
		    size_t n)
{
	int b;
	int i;
	int e;

	/* Find beginning (b): first non-whitespace character. */
	b = 0;
	while ((b < n) && isspace(src[b])) {
		b++;
	}

	/* Find printables, remembering last non-space character at e. */
	e = i = b;
	while ((i < n) && isprint(src[i])) {
		if (!isspace(src[i]))
			e = i;
		i++;
	}

	/* Make e point past last useful character. */
	e++;

	/* Duplicate string between b and e. */
	if (e >= b) {
		memmove(dest, &src[b], e - b);
		dest[e - b] = '\0';
	}

	return dest;
}

/* This potentially wastes memory, but not much... */
char *
strndup_trim(const unsigned char * s,
	     size_t n)
{
	char * ret;

	ret = malloc(n + 1);

	if (NULL != ret)
		lsvpd_strnmove_trim(ret, s, n);

	return ret;
}

char *
lsvpd_hexify(const unsigned char * s,
	     size_t len)
{
	int i;
	char * ret = malloc(len*2+1);
	for (i = 0; i < len ; i++) {
		snprintf(&ret[i*2], 3, "%.2X", s[i]);
	}

	return ret;
}

void
lsvpd_strtolower(char * s)
{
	while ('\0' != *s) {
		if (isupper(*s))
			*s = tolower(*s);
		s++;
	}
}

bool
lsvpd_prefix_substitute(char ** s,
			const char * from,
			const char * to)
{
	bool ret;

	size_t len;
	char * t;

	ret = false;

	len = strlen(from);

	if ((0 == strncmp(from, *s, len)) &&
	    (-1 != asprintf(&t, "%s%s", to, *s + len))) {
		free(*s);
		*s = t;
		ret = true;
	}

	return ret;
}

bool
line_scanner_harness(const char * filename,
		     bool (* line_scanner) (const char *, void *),
		     void * result)
{
	bool ret;
	unsigned char * data;
	size_t len;
	unsigned char * this_line;
	unsigned char * next_line;
	bool more;

	ret = false;

	if (lsvpd_file_read(filename, &data, &len)) {
		more = true;
		this_line = data;
		while (more && (NULL != this_line)) {
			next_line = memchr(this_line, '\n',
					   len - (this_line - data));
			if (NULL != next_line) {
				*next_line = '\0';
				next_line++;
			}

			more = (*line_scanner)((char *) this_line, result);

			if ((NULL != next_line) && ((next_line - data) < len))
				this_line = next_line;
			else
				this_line= NULL;
		}
		free(data);

		ret = true;
		
	}

	return ret;
}

/********************************************************************/

/* Pass the name of a lock file here, not the name of the file you
 * want to lock .  */

int
lsvpd_lock(const char * f)
{
	int ret;
	struct flock fl;

	ret = creat(f, O_CREAT);
	if (-1 != ret) {
		fl.l_type = F_WRLCK;
		fl.l_whence = SEEK_SET;
		fl.l_start = 0;
		fl.l_len = 0;

		if (-1 == fcntl(ret, F_SETLKW, &fl)) {
			close(ret);
			ret = -1;
		}
	}

	return ret;
}

void
lsvpd_unlock(int fd)
{
	/* Closing file removes locks. */
	close(fd);
}

